from __future__ import annotations

import json
from abc import ABC, abstractmethod
from typing import Any, Enum, NamedTuple

NO_DEFAULT = object()


class EncodableField(ABC):
    def __init__(self, data_field, default=NO_DEFAULT, transition: Transition | None = None):
        self.data_field = data_field
        self.default = default
        self.transition = transition

    @abstractmethod
    @property
    def name(self):
        ...

    def check_data(self, data):
        return data

    def to_json(self):
        specs = {'name': self.name, 'field': self.data_field}
        if self.default is not NO_DEFAULT:
            specs['default'] = self.default
        if self.transition is not None:
            specs['transition'] = self.transition.to_json()
        return json.dumps(specs, ensure_ascii=True)


class TransitionEasing(str, Enum):
    LINEAR = 'linear'
    EASE_IN = 'easeIn'
    EASE_OUT = 'easeOut'
    EASE_IN_OUT = 'easeInOut'


class Transition(NamedTuple):
    duration: int = 0
    easing: TransitionEasing = TransitionEasing.LINEAR

    def to_json(self):
        return json.dumps({'easing': self.easing.value, 'duration': self.duration}, ensure_ascii=True)


class ScaleType(str, Enum):
    LINEAR = 'linear'
    LOG = 'log'
    SQRT = 'sqrt'
    QUANTILE = 'quantile'


class Scale(NamedTuple):
    type: ScaleType
    domain: tuple[Any, ...]
    clamp: bool = False

class PowScale(Scale):
    type: str = 'pow'



class FieldValueError(Exception):
    def __init__(self, field: EncodableField, value):
        self.field = field
        self.value = value

    def __str__(self):
        return f"Invalid value for field {self.field.name}: {self.value} (from data column: {self.field.data_field})."

#############################################
class NumberEncodableField(EncodableField):
    def check_data(self, data):
        try:
            return float(data)
        except ValueError:
            raise FieldValueError(self, data) from None


class ColorEncodableField(EncodableField):
    def check_data(self, data):
        import webcolors
        if isinstance(data, str):
            try:
                return webcolors.name_to_rgb(data)
            except ValueError:
                raise FieldValueError(self, data) from None

#############################################
class X(NumberEncodableField):
    @property
    def name(self):
        return "x"


class Y(NumberEncodableField):
    @property
    def name(self):
        return "y"


class Width(NumberEncodableField):
    @property
    def name(self):
        return "width"


class Height(NumberEncodableField):
    @property
    def name(self):
        return "height"


class Color(ColorEncodableField):
    @property
    def name(self):
        return "color"



import os
import re
from pathlib import Path
from typing import Tuple

import cv2
import numpy as np


def load_image_from_url(url: str | Path):
    if isinstance(url, Path):
        url = str(url)
    if os.path.exists(url):
        img = cv2.imread(url)
        if img is None:
            raise ValueError(f"Invalid image file: {url}.") from None
    elif re.match(r"^https?://", url):
        from urllib.request import urlopen

        resp = urlopen(url)
        img = np.asarray(bytearray(resp.read()), dtype=np.uint8)
        img = cv2.imdecode(img, -1)
        if img is None:
            raise ValueError(f"Invalid image url: {url}.")
    else:
        raise ValueError(f"No file was found at: {url}.")
    
    return img


def fit_resize(img: np.ndarray, size: Tuple[int, int] | int, interpolation=None) -> np.ndarray:
        if isinstance(size, int):
            size = (size, size)

        # Keep aspect ratio
        h, w = img.shape[:2]
        ratio = h / w
        mindim = min(size[0] * ratio, size[1])
        size = (round(mindim / ratio), round(mindim))

        # Select interpolation method based on image resize
        if interpolation is None:
            if size[0] > w or size[1] > h:
                interpolation = cv2.INTER_CUBIC
            else:
                interpolation = cv2.INTER_AREA

        # Resize image
        return cv2.resize(img, size, interpolation=interpolation)


from __future__ import annotations

from functools import reduce
from typing import Iterable, List, NamedTuple, Tuple, TypeGuard, overload

import numpy as np
import numpy.typing as npt

from .func import StrEnum


class Transform:
    def __init__(self, translate: Point = (0, 0), scale: float = 1, origin: Point = (0, 0)):
        self._translate = translate
        self._scale = scale
        self._origin = origin

    def __repr__(self):
        return f"Transform(translate={self._translate}, scale={self._scale}, origin={self._origin})"

    @overload
    def __call__(self, p: tuple) -> tuple:
        ...

    @overload
    def __call__(self, p: Point) -> Point:
        ...

    @overload
    def __call__(self, p: Rect) -> Rect:
        ...

    def __call__(self, p: tuple | Point | Rect) -> tuple | Point | Rect:
        if isinstance(p, tuple):
            match len(p):
                case 2:
                    p = Point(*p)
                case 4:
                    p = Rect(*p)
                case _:
                    raise TypeError(
                        "Only transformation of point (2-items tuple) or rect (4-items tuple) are supported"
                    )
        if isinstance(p, Rect):
            return Rect.from_points(self(p.top_left), self(p.bottom_right))
        elif isinstance(p, Point):
            return ((p - self._origin) * self._scale) + self._origin + self._translate
        else:
            raise TypeError("Only Rect and Point are supported")

    @property
    def translate(self) -> Point:
        return self._translate

    @property
    def scale(self) -> float:
        return self._scale

    @property
    def origin(self) -> Point:
        return self._origin

    @staticmethod
    def from_rects(src: Rect, dst: Rect):
        origin = Rect(*src)
        target = Rect(*dst)

        offset = target.top_left - origin.top_left
        if origin.w != 0:
            ratio = target.w / origin.w
        elif origin.h != 0:
            ratio = target.h / origin.h
        else:
            raise ValueError("Origin rect cannot be empty")
        return Transform(offset, ratio, origin.top_left)


class FitMode(StrEnum):
    WIDTH = "fit_width"
    HEIGHT = "fit_height"
    INNER = "fit_inner"
    OUTER = "fit_outer"
    CENTERED = "centered"


class Rect(NamedTuple):
    h: float
    w: float
    y: float = 0
    x: float = 0

    @property
    def center(self) -> Point:
        return Point(self.y + self.h // 2, self.x + self.w // 2)

    @property
    def top_left(self) -> Point:
        return Point(self.y, self.x)

    @property
    def bottom_right(self) -> Point:
        return Point(self.y + self.h, self.x + self.w)

    @property
    def top(self) -> float:
        return self.y

    @property
    def bottom(self) -> float:
        return self.y + self.h

    @property
    def left(self) -> float:
        return self.x

    @property
    def right(self) -> float:
        return self.x + self.w

    @property
    def shape(self) -> Point:
        return Point(y=self.h, x=self.w)

    @property
    def area(self) -> float:
        return self.h * self.w

    def to_int(self):
        return Rect(*(int(round(_)) for _ in (self.h, self.w, self.y, self.x)))

    @classmethod
    def from_tuple(
        cls,
        rect: float
        | int
        | Tuple[float | int]
        | Tuple[float | int, float | int]
        | Tuple[float | int, float | int, float | int, float | int],
    ):
        match rect:
            case size if isinstance(size, (float, int)):
                rect = (size, size)
            case (h, w) if all(isinstance(_, (float, int)) for _ in rect):
                rect = (h, w)
            case (h, w, y, x) if all(isinstance(_, (float, int)) for _ in rect):
                rect = (h, w, y, x)
            case _:
                raise TypeError("Rect can only be created from a float or a tuple of 2 or 4 floats")
        return cls(*rect)

    @classmethod
    def from_size(cls, shape: Tuple[float | int, float | int]):
        return cls(shape[0], shape[1])

    @overload
    @classmethod
    def from_points(cls, bottom_right: Tuple[float | int, float | int]) -> Rect:
        ...

    @overload
    @classmethod
    def from_points(cls, bottom: float | int, right: float | int) -> Rect:
        ...

    @overload
    @classmethod
    def from_points(
        cls,
        top_left: Tuple[float | int, float | int],
        bottom_right: Tuple[float | int, float | int],
        *,
        ensure_positive: bool,
    ) -> Rect:
        ...

    @overload
    @classmethod
    def from_points(
        cls, top: float | int, left: float | int, bottom: float | int, right: float | int, *, ensure_positive: bool
    ) -> Rect:
        ...

    @overload
    @classmethod
    def from_points(
        cls, top_left_bottom_right: Tuple[float | int, float | int, float | int, float | int], *, ensure_positive: bool
    ) -> Rect:
        ...

    @classmethod
    def from_points(
        cls,
        *p: float | int | Tuple[float | int, float | int] | Tuple[float | int, float | int, float | int, float | int],
        ensure_positive: bool = False,
    ) -> Rect:
        match p:
            case ((bottom, right),) if all(isinstance(_, float | int) for _ in p[0]):
                p2 = Point(bottom, right)
                p1 = Point.origin()
            case (bottom, right) if all(isinstance(_, float | int) for _ in p):
                p2 = Point(bottom, right)
                p1 = Point.origin()
            case ((top, left), (bottom, right)) if all(isinstance(_, float | int) for _ in p[0] + p[1]):
                p1 = Point(top, left)
                p2 = Point(bottom, right)
            case (top, left, bottom, right) if all(isinstance(_, float | int) for _ in p):
                p1 = Point(top, left)
                p2 = Point(bottom, right)
            case ((top, left, bottom, right),) if all(isinstance(_, float | int) for _ in p[0]):
                p1 = Point(top, left)
                p2 = Point(bottom, right)
            case _:
                raise TypeError("Rect can only be created from 2 or 4 floats or from 2 tuples of 2 floats")

        if not ensure_positive:
            return cls(abs(p2.y - p1.y), abs(p2.x - p1.x), min(p1.y, p2.y), min(p1.x, p2.x))
        else:
            rect = cls(p2.y - p1.y, p2.x - p1.x, p1.y, p1.x)
            return Rect.empty() if rect.h < 0 or rect.w < 0 else rect

    @classmethod
    def from_center(cls, center: Tuple[float, float], shape: float | Tuple[float, float]) -> Rect:
        if isinstance(shape, (float, int)):
            shape = (shape, shape)
        return cls(shape[0], shape[1], center[0] - shape[0] // 2, center[1] - shape[1] // 2)

    @classmethod
    def empty(cls) -> Rect:
        return cls(0, 0, 0, 0)

    def is_self_empty(self) -> bool:
        return self.w == 0 or self.h == 0

    @classmethod
    def is_empty(cls, rect: Rect | None) -> bool:
        if rect is None:
            return True
        if isinstance(rect, tuple) and len(rect) == 4:
            rect = Rect(*rect)
        return isinstance(rect, tuple) and (rect.w == 0 or rect.h == 0)

    @classmethod
    def is_rect(cls, r) -> TypeGuard[Rect]:
        return isinstance(r, Rect) or (isinstance(r, tuple) and len(r) == 4)

    def __repr__(self):
        return "Rect(y={}, x={}, h={}, w={})".format(self.y, self.x, self.h, self.w)

    def __or__(self, other) -> Rect:
        if isinstance(other, Rect):
            if self.is_self_empty():
                return other
            if other.is_self_empty():
                return self
            return Rect.from_points(
                (min(self.top, other.top), min(self.left, other.left)),
                (max(self.bottom, other.bottom), max(self.right, other.right)),
            )
        else:
            raise TypeError("Rect can only be combined only with another Rect")

    def __and__(self, other) -> Rect:
        if isinstance(other, Rect):
            return Rect.from_points(
                (max(self.top, other.top), max(self.left, other.left)),
                (min(self.bottom, other.bottom), min(self.right, other.right)),
                ensure_positive=True,
            )
        else:
            raise TypeError("Rect can only be combined only with another Rect")

    def __bool__(self) -> bool:
        return not self.is_self_empty()

    def __add__(self, other: Point | float) -> Rect:
        if isinstance(other, float):
            other = Point(other, other)
        if isinstance(other, Point):
            return self.translate(other.y, other.x)
        raise TypeError("Rect can only be translated by a Point or a float")

    def __sub__(self, other: Point | float) -> Rect:
        if isinstance(other, float):
            other = Point(other, other)
        if isinstance(other, Point):
            return self.translate(-other.y, -other.x)
        raise TypeError("Rect can only be translated by a Point or a float")

    def __mul__(self, other: float) -> Rect:
        return self.scale(other)

    def __truediv__(self, other: float) -> Rect:
        return self.scale(1 / other)

    def __contains__(self, other: Point | Rect) -> bool:
        if isinstance(other, Point):
            return self.y <= other.y <= self.y + self.h and self.x <= other.x <= self.x + self.w
        elif isinstance(other, Rect):
            return not Rect.is_empty(self & other)
        else:
            raise TypeError("Rect can only be compared with a Point or a Rect")

    def translate(self, y: float, x: float) -> Rect:
        return Rect(self.h, self.w, self.y + y, self.x + x)

    def scale(self, fy: float, fx: float | None = None) -> Rect:
        if fx is None:
            fx = fy
        return Rect(self.h * fy, self.w * fx, self.y * fy, self.x * fx)

    def fit(self, other: Rect | Point | tuple, mode: FitMode = FitMode.WIDTH) -> Rect:
        match other:
            case (h, w):
                other = Rect.from_size((h, w))
            case (h, w, y, x):
                other = Rect(h, w, y, x)
        match mode:
            case FitMode.OUTER:
                ratio = max(other.w / self.w, other.h / self.h)
            case FitMode.INNER:
                ratio = min(other.w / self.w, other.h / self.h)
            case FitMode.WIDTH:
                ratio = other.w / self.w
            case FitMode.HEIGHT:
                ratio = other.h / self.h
            case _:
                ratio = 1
        return Rect.from_center(other.center, (self.h * ratio, self.w * ratio))

    def transforms_to(self, target: Rect) -> Transform:
        return Transform.from_rects(self, target)

    def clip(self, rect: float | Tuple[float, float] | Tuple[float, float, float, float]) -> Rect:
        rect = Rect.from_tuple(rect)
        return Rect.from_points(
            (max(self.top, rect.top), max(self.left, rect.left)),
            (min(self.bottom, rect.bottom), min(self.right, rect.right)),
            ensure_positive=True,
        )

    @overload
    def pad(self, pad: float | Tuple[float, float]) -> Rect:
        ...

    @overload
    def pad(self, vertical: float, horizontal: float) -> Rect:
        ...

    @overload
    def pad(self, top: float, right: float, bottom: float, left: float) -> Rect:
        ...

    def pad(self, *pad: float | Tuple[float, float]) -> Rect:
        match pad:
            case (pad,) if isinstance(pad, int | float):
                pad = (pad,) * 4
            case ((vertical, horizontal),) if all(isinstance(_, int | float) for _ in pad[0]):
                pad = (vertical, horizontal) * 2
            case (pad_top, pad_right, pad_bottom, pad_left) if all(isinstance(_, int) for _ in pad):
                pass
            case _:
                raise ValueError("pad() accepts 1, 2 or 4 arguments")
        return Rect(self.h + pad[0] + pad[2], self.w + pad[1] + pad[3], self.y - pad[0], self.x - pad[3])

    def slice(self) -> tuple[slice, slice]:
        r = self.to_int()
        return slice(r.y, r.y + r.h), slice(r.x, r.x + r.w)

    @staticmethod
    def union(*rects: Tuple[Iterable[Rect] | Rect, ...]) -> Rect:
        rects = sum(((r,) if isinstance(r, Rect) else tuple(r) for r in rects), ())
        return reduce(lambda a, b: a | b, rects)

    @staticmethod
    def intersection(*rects: Tuple[Iterable[Rect] | Rect, ...]) -> Rect:
        rects = sum(((r,) if isinstance(r, Rect) else tuple(r) for r in rects), ())
        return reduce(lambda a, b: a & b, rects)


class Point(NamedTuple):
    y: float
    x: float

    def xy(self) -> tuple[float, float]:
        return self.x, self.y

    def to_int(self) -> Point:
        return Point(int(round(self.y)), int(round(self.x)))

    def clip(self, rect: float | Tuple[float, float] | Tuple[float, float, float, float]) -> Point:
        rect = Rect.from_tuple(rect)
        return Point(
            min(max(self.y, rect.top), rect.bottom),
            min(max(self.x, rect.left), rect.right),
        )

    def __add__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(self.y + other, self.x + other)
        y, x = other
        return Point(self.y + y, self.x + x)

    def __radd__(self, other: Tuple[float, float] | float):
        return self + other

    def __sub__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(self.y - other, self.x - other)
        y, x = other
        return Point(self.y - y, self.x - x)

    def __rsub__(self, other: Tuple[float, float] | float):
        return -(self - other)

    def __mul__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(self.y * other, self.x * other)
        y, x = other
        return Point(self.y * y, self.x * x)

    def __rmul__(self, other: Tuple[float, float] | float):
        return self * other

    def __truediv__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(self.y / other, self.x / other)
        y, x = other
        return Point(self.y / y, self.x / x)

    def __rtruediv__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(other / self.y, other / self.x)
        y, x = other
        return Point(y / self.y, x / self.x)

    def __neg__(self):
        return Point(-self.y, -self.x)

    @classmethod
    def origin(cls):
        return cls(0, 0)

    @classmethod
    def from_tuple(cls, point: float | int | Tuple[float | int] | Tuple[float | int, float | int]):
        match point:
            case (y, x) if all(isinstance(_, (float, int)) for _ in point):
                return cls(y, x)
            case size if isinstance(size, (float, int)):
                return cls(size, size)
            case (size,) if isinstance(size, (float, int)):
                return cls(size, size)
            case _:
                raise TypeError("Point can only be created from a float or a tuple of 2 floats")

    @overload
    def distance(self, other: Point) -> float:
        ...

    @overload
    def distance(self, other: List[Point]) -> List[float]:
        ...

    @overload
    def distance(self, other: npt.NDArray[np.float]) -> npt.NDArray[np.float]:
        ...

    def distance(self, other: Point | Iterable[Point]) -> float | Iterable[float]:
        import numpy as np

        if isinstance(other, np.ndarray):
            return np.linalg.norm(other - self, axis=-1)
        elif isinstance(other, list):
            return [self.distance(p) for p in other]
        return ((self.y - other.y) ** 2 + (self.x - other.x) ** 2) ** 0.5



import inspect
from enum import EnumMeta
from typing import Mapping, Protocol

from strenum import StrEnum as StrEnumBase


class FlagContextSetFlag(Protocol):
    def __call__(self, flag_value: bool):
        ...


class FlagContext:
    def __init__(self, set_flag: FlagContextSetFlag, value_on_enter=True):
        self.set_flag = set_flag
        self.value_on_enter = value_on_enter
        self.enter_count = 0

    def __enter__(self):
        if self.enter_count == 0:
            self.set_flag(self.value_on_enter)
        self.enter_count += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.enter_count -= 1
        if self.enter_count == 0:
            self.set_flag(not self.value_on_enter)


def call_matching_params(method, args=None, kwargs=None):
    """
    Call the specified method, matching the arguments it needs with those,
    provided in kwargs. The useless arguments are ignored.
    If some not optional arguments is missing, a ValueError exception is raised.
    :param method: The method to call
    :param kwargs: Parameters provided to method
    :return: Whatever is returned by method (might be None)
    """
    method_params = {}
    if kwargs:
        method_params = inspect.signature(method).parameters.keys()
        method_params = {_: kwargs[_] for _ in method_params & kwargs.keys()}

    if args is None:
        args = []
    i_args = 0
    for not_opt in not_optional_args(method):
        if not_opt not in method_params:
            if i_args < len(args):
                method_params[not_opt] = args[i_args]
                i_args += 1
            else:
                raise ValueError("%s is not optional to call method: %s." % (not_opt, method))

    return method(**method_params)


def not_optional_args(f):
    """
    List all the parameters not optional of a method
    :param f: The method to analise
    :return: The list of parameters
    :rtype: list
    """
    sig = inspect.signature(f)
    return [p_name for p_name, p in sig.parameters.items() if p.default is inspect.Parameter.empty]


def dict_recursive_update(d1: dict, d2: Mapping) -> Mapping:
    for k, v in d2.items():
        if isinstance(v, Mapping):
            d1[k] = dict_recursive_update(d1.get(k, {}), v)
        else:
            d1[k] = v
    return d1


class MetaEnum(EnumMeta):
    def __contains__(cls, item):
        return item in tuple(cls)


class StrEnum(StrEnumBase, metaclass=MetaEnum):
    ...



from functools import reduce

from IPython.display import display
from ipywidgets import HTML


def vscode_theme(ipywidget=True):
    # Redefine root variables colors to match VSCode theme
    root_vars = {
        "jppype-foreground-color": ("vscode-editor-foreground", "jp-content-font-color1"),
        "jppype-dim-foreground-color": ("vscode-input-placeholderForeground", "jp-content-font-color2"),
        "jppype-dim-background-color": ("vscode-input-background", "jp-layout-color3"),
        "jppype-background-color": ("vscode-editor-background", "jp-layout-color2"),
        "jppype-highlight-background-color": ("vscode-panel-background", "jp-layout-color1"),
    }
    if ipywidget:
        root_vars["jp-widgets-color"] = root_vars["jppype-foreground-color"]

        root_vars["jp-widgets-slider-handle-border-color"] = (("vscode-input-border", "jp-border-color"),)
        root_vars["jp-widgets-slider-handle-background-color"] = (
            "vscode-scrollbarSlider-background",
            "jp-layout-color1",
        )
        root_vars["jp-widgets-slider-active-handle-color"] = ("vscode-focusBorder", "jp-brand-color1")

        root_vars["jp-widgets-input-color"] = ("vscode-input-foreground", "jp-ui-font-color1")
        root_vars["jp-widgets-input-background-color"] = ("vscode-input-background", "jp-layout-color1")
        root_vars["jp-widgets-input-border-color"] = ("vscode-input-border", "jp-border-color")
        root_vars["jp-widgets-slider-active-handle-color"] = ("vscode-inputOption-activeForeground", "jp-brand-color1")
        root_vars["jp-widgets-input-focus-border-color"] = ("vscode-inputOption-activeBorder", "jp-brand-color2")

    root_vars_css = "\n".join(
        f"    --{var_name}: "
        + reduce(lambda x, y: f"var(--{y}, {x})", reversed(var_values[:-1]), f"var(--{var_values[-1]})")
        + ";"
        for var_name, var_values in root_vars.items()
    )

    # Background disabling
    background_css = ""
    if ipywidget:
        background_css = """
        .cell-output-ipywidget-background {
                background: transparent !important;
        }"""
    else:
        background_css = """
        .jbasewidget {
            background: var(--vscode-editor-background, transparent) !important;
        }"""

    theme = (
        "<style>"
        + background_css
        + """
        :root {"""
        + root_vars_css
        + """
        }
        </style>"""
    )

    display(HTML(theme))



import json

from IPython import display
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from yaml import Loader, load


class ChangeHandler(FileSystemEventHandler):
    def __init__(self, vega):
        self.vega = vega

    def on_modified(self, event):
        self.vega.update()


class Vega:
    def __init__(self, spec):
        self._spec = spec
        if isinstance(spec, str):
            change_handler = ChangeHandler(self)
            self.observer = Observer()
            self.observer.schedule(change_handler, path=spec)
            self.observer.start()

    def __delete__(self):
        self.observer.stop()
        self.observer.join()

    def json(self):
        return json.dumps(self.spec, indent=2)

    @property
    def spec(self):
        if isinstance(self._spec, str):
            if self._spec.endswith(".yaml"):
                spec = load(open(self._spec), Loader=Loader)
            elif self._spec.endswith(".json"):
                spec = json.load(open(self._spec))
        else:
            spec = self._spec
        return spec

    def repr_html(self):
        return f"""
        <div id="vis"></div>
        <script>
        var spec = {self.spec};
        vegaEmbed('#vis', spec);
        </script>
        """

    def _repr_mimebundle_(self, include, exclude):
        return {
            "application/vnd.vega.v5+json": self.json(),
            "text/html": self.repr_html(),
        }

    def update(self):
        display.clear_output(wait=True)
        display.display(self)


from __future__ import annotations
from typing import TypeVar, Generic, Protocol, Literal, Iterable
from asyncio import Future


class Event(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


E = TypeVar("E", bound=Event)


class EventCallBack(Generic[E], Protocol):
    def __call__(self, event: E):
        ...


class EventsDispatcherCallback(Generic[E]):
    def __init__(self, cb: EventCallBack[E], once: bool = False):
        self._cb = cb
        self._once = once
        self._dispatcher = []

    def __call__(self, event: E):
        if self._once:
            self.unsubscribe()
        self._cb(event)

    def _aknowlowdge_subscribe(self, dispatcher: EventsDispatcher):
        self._dispatcher.append(dispatcher)

    def unsubscribe(self):
        for d in self._dispatcher:
            d.unsubscribe(self)
        self._dispatcher.clear()


class EventFuture(EventsDispatcherCallback[E], Future):
    def __init__(self):
        EventsDispatcherCallback.__init__(self, self._on_event, True)
        Future.__init__(self)

    def _on_event(self, event: E):
        if self.done:
            if self.cancelled():
                print("EventFuture: cancelled")
                print(self.exception())
            else:
                print("EventFuture: already done")
                print(self.result())
        else:
            self.set_result(event)

    def result(self) -> E:
        return Future.result(self)


class EventsDispatcher(Generic[E]):
    def __init__(self):
        self._cb = []

    def __call__(self, cb: EventCallBack[E], once=False) -> EventsDispatcherCallback:
        return self.subscribe(cb, once)

    def dispatch(self, event: E):
        for cb in self._cb:
            cb(event)

    def create_future(self) -> EventFuture[E]:
        cb = EventFuture()
        self._cb.append(cb)
        cb._aknowlowdge_subscribe(self)
        return cb

    def subscribe(self, cb: EventCallBack[E], once=False):
        cb = EventsDispatcherCallback(cb, once)
        self._cb.append(cb)
        cb._aknowlowdge_subscribe(self)
        return cb

    def unsubscribe(self, cb: EventsDispatcherCallback):
        self._cb.remove(cb)


###################################################################################################
#           Events Specialization                                                                 #
###################################################################################################

MouseModifier = Literal["alt", "control", "shift", "meta"]


class ClickEvent(Event):
    def __init__(self, x: float, y: float, button: int, modifiers: Iterable[MouseModifier]):
        super().__init__(**{"x": x, "y": y, "button": button, "modifiers": modifiers})
        self.x = x
        self.y = y
        self.button = button
        self.modifiers = modifiers



from __future__ import annotations

import abc
from copy import copy
from typing import (
    Callable,
    Dict,
    Iterable,
    Literal,
    Mapping,
    Protocol,
    Tuple,
    Type,
    TypeGuard,
    get_args,
)
from uuid import uuid4

from ..utilities.func import call_matching_params
from ..utilities.geometric import FitMode, Rect, Transform


# ======================================================================================================================
#   Utility class used by Layer and LayersList
# ======================================================================================================================
class LayerData:
    def __init__(self, data: any, infos: dict = None, type: str = None):
        self.data = data
        self.type = type
        self.infos = infos

    def to_json_bytes(self) -> bytes:
        import json

        return json.dumps(
            dict(data=self.data, type=self.type, infos=self.infos),
            indent=None,
            separators=(",", ":"),
            ensure_ascii=True,
        ).encode("ascii")


class LayerDataChangeDispatcher(Protocol):
    def __call__(self, layer: Layer):
        ...


class LayerOptionsChangeDispatcher(Protocol):
    def __call__(self, layer: Layer, options: Dict[str, any]):
        ...


class DispatcherUnbind:
    def __init__(self, dispatchers_dict, uuid):
        self._dispatchers_dict = dispatchers_dict
        self._uuid = uuid

    def __call__(self):
        try:
            del self._dispatchers_dict[self._uuid]
        except KeyError:
            pass


class LayerSelector(Protocol):
    def __call__(self, name: str, layer: Layer) -> bool:
        ...


class ContextLock:
    def __init__(self, final_callback: Callable[[dict[str, set]], None]):
        self._locked = False
        self._flags: dict[str, set] = {}
        self._final_callback = final_callback
        self._enter_count = 0

    def __enter__(self):
        self._locked = True
        self._enter_count += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._enter_count -= 1

        if self._enter_count == 0:
            self._final_callback(self._flags)
            self._flags.clear()
            self._locked = False

    def add(self, flag: str, data=None):
        if self._locked:
            if flag not in self._flags:
                self._flags[flag] = set()
            if data is not None:
                self._flags[flag].add(data)
        return self

    def __contains__(self, item: str):
        return item in self._flags

    @property
    def flags(self) -> dict[str, set]:
        return self._flags

    def __bool__(self):
        return self._locked

    @property
    def locked(self) -> bool:
        return self._locked


LayerDomain = Rect | FitMode
DomainMode = Literal["manual"] | FitMode


def is_domain(value: any) -> TypeGuard[LayerDomain]:
    return isinstance(value, (Rect, str)) or value in FitMode


# ======================================================================================================================
#   Layer base class
# ======================================================================================================================


BlendMode = Literal[
    "normal",
    "multiply",
    "screen",
    "overlay",
    "darken",
    "lighten",
    "color_dodge",
    "color_burn",
    "hard_light",
    "soft_light",
    "difference",
    "exclusion",
    "hue",
    "saturation",
    "color",
    "luminosity",
]


class Layer(abc.ABC):
    def __init__(self, layer_type: str):
        self._options = {
            "visible": True,
            "opacity": 1.0,
            "z_index": -1,
            "foreground": False,
            "label": "",
            "auto_scale_domain": True,
            "domain": Rect.empty(),
        }
        self._layer_type = layer_type
        self._on_data_change: Dict[str, LayerDataChangeDispatcher] = {}
        self._on_options_change: Dict[str, LayerOptionsChangeDispatcher] = {}
        self._main_domain = Rect.empty()
        self._domain_mode: DomainMode | None = None
        self._uuid = uuid4()

    def duplicate(self):
        layer = copy(self)
        layer._uuid = uuid4()
        return layer

    # --- Base properties ---
    @property
    def uuid(self) -> str:
        return self._uuid.hex

    def __hash__(self):
        return hash(self._uuid)

    def __eq__(self, other):
        return isinstance(other, Layer) and self._uuid == other._uuid

    @property
    def layer_type(self) -> str:
        return self._layer_type

    @property
    def options(self) -> Mapping[str, any]:
        return self._options

    @property
    def shape(self) -> Tuple[int, int]:
        return self._shape()

    # --- Options properties ---
    def set_options(self, options: Dict[str, any], raise_on_error: bool = True):
        for k, v in options.items():
            match k:
                case "visible":
                    if not isinstance(v, bool):
                        if raise_on_error:
                            raise ValueError(f"visible must be a bool, got {v}")
                        else:
                            continue
                    self._options["visible"] = v

                case "opacity":
                    if not isinstance(v, (int, float)):
                        try:
                            v = float(v)
                        except ValueError:
                            if raise_on_error:
                                raise ValueError(f"opacity must be a number between 0 and 1, got {v}") from None
                            else:
                                continue
                    if not 0 <= v <= 1:
                        if raise_on_error:
                            raise ValueError(f"opacity must be a number between 0 and 1, got {v}")
                        else:
                            continue
                    self._options["opacity"] = v

                case "blend_mode":
                    if v not in get_args(BlendMode):
                        if raise_on_error:
                            raise ValueError(f"blend_mode must be one of {BlendMode}, got {v}")
                        else:
                            continue
                    self._options["blend_mode"] = v

                case "label":
                    if not isinstance(v, str):
                        if raise_on_error:
                            raise ValueError(f"label must be a string, got {v}")
                        else:
                            continue
                    self._options["label"] = v

                case "z_index":
                    if not isinstance(v, (int, float)):
                        try:
                            v = float(v)
                        except ValueError:
                            if raise_on_error:
                                raise ValueError(f"z_index must be a number, got {v}") from None
                            else:
                                continue
                    self._options["z_index"] = v
                case "domain":
                    if (
                        not isinstance(v, (tuple, list))
                        or len(v) != 4
                        or not all(isinstance(x, (int, float)) for x in v)
                    ):
                        if raise_on_error:
                            raise (ValueError(f"domain must be a tuple or list of 4 numbers, got {v}"))
                        else:
                            continue
                    self._options["domain"] = tuple(v)
                case "foreground":
                    if not isinstance(v, bool):
                        if raise_on_error:
                            raise ValueError(f"foreground must be a bool, got {v}")
                        else:
                            continue
                    self._options["foreground"] = v
        self._notify_options_change({k: self._options[k] for k in options.keys() if k in self._options})

    @property
    def visible(self) -> bool:
        return self._options["visible"]

    @visible.setter
    def visible(self, value: bool):
        self.set_options({"visible": value})

    @property
    def opacity(self) -> float:
        return self._options["opacity"]

    @opacity.setter
    def opacity(self, value: float):
        self.set_options({"opacity": value})

    @property
    def blend_mode(self) -> BlendMode:
        return self._options["blend_mode"]

    @blend_mode.setter
    def blend_mode(self, value: BlendMode):
        self.set_options({"blend_mode": value})

    @property
    def label(self) -> str:
        return self._options["label"]

    @label.setter
    def label(self, value: str):
        self.set_options({"label": value})

    @property
    def z_index(self) -> float:
        return self._options["z_index"]

    @z_index.setter
    def z_index(self, value: float):
        self.set_options({"z_index": value})

    @property
    def foreground(self) -> bool:
        return self._options["foreground"]

    @foreground.setter
    def foreground(self, value: bool):
        self.set_options({"foreground": value})

    @property
    def domain(self) -> Rect:
        return Rect(*self._options["domain"])

    @domain.setter
    def domain(self, value: LayerDomain | None):
        shape = Rect.from_size(self.shape)
        match value:
            case value if value is None or Rect.is_empty(value):
                if not Rect.is_empty(shape):
                    value = Rect.from_size(self.shape).fit(self._main_domain, FitMode.WIDTH)
                    self._domain_mode = "manual"
            case v if v in FitMode:
                if not Rect.is_empty(shape):
                    value = Rect.from_size(self.shape).fit(self._main_domain, value)
                    self._domain_mode = value
            case _:
                value = Rect(*value)
        self.set_options({"domain": value})

    @property
    def domain_mode(self) -> FitMode | None:
        return self._domain_mode

    @domain_mode.setter
    def domain_mode(self, value: FitMode):
        raise AttributeError("domain_mode is read-only, use domain setter instead.")

    def set_main_shape(self, main_domain: Rect, transform_domain: Transform | LayerDomain | None = None):
        previous_domain = self._main_domain
        self._main_domain = main_domain

        match self.domain_mode:
            case "manual" | None:
                match transform_domain:
                    case None:
                        if main_domain is not None:
                            transform_domain = Transform.from_rects(previous_domain, main_domain)
                            self.domain = transform_domain(self.domain)
                    case Transform():
                        self.domain = transform_domain(self.domain)
                    case transform_domain if is_domain(transform_domain):
                        self.domain = transform_domain
            case v if v in FitMode:
                self.domain = self.domain_mode

    # --- Fetch data methods ---
    def get_data(self, **kwargs) -> LayerData:
        r = call_matching_params(self._fetch_data, kwargs)
        if isinstance(r, bytes):
            r = LayerData(r)
        if isinstance(r, LayerData):
            if r.type is None:
                r.type = self._layer_type
            return r
        raise TypeError(f"Invalid return type {type(r)}")

    def fetch_item(self, **kwargs) -> dict:
        return call_matching_params(self._fetch_item, kwargs)

    def fetch_graphs(self, rect: Tuple[float, float], **kwargs) -> Dict[str | str]:
        kwargs["rect"] = rect
        return call_matching_params(self._fetch_graphs, kwargs)

    # --- Abstract methods ---
    @abc.abstractmethod
    def _fetch_data(self, **kwargs) -> LayerData:
        ...

    @abc.abstractmethod
    def update_data(self, data: any):
        ...

    @abc.abstractmethod
    def _fetch_item(self, **kwargs) -> Mapping[str, any]:
        ...

    @abc.abstractmethod
    def _fetch_graphs(self, rect: Tuple[float, float], **kwargs) -> Dict[str, any]:
        ...

    @abc.abstractmethod
    def _shape(self) -> Tuple[int, int]:
        ...

    # --- Communication method ---
    def _notify_data_change(self):
        for callback in self._on_data_change.values():
            callback(self)

    def on_data_change(self, callback: LayerDataChangeDispatcher):
        uuid = uuid4().hex
        while uuid in self._on_data_change:
            uuid = uuid4().hex
        self._on_data_change[uuid] = callback
        return DispatcherUnbind(self._on_data_change, uuid)

    def _notify_options_change(self, options_changed: Dict[str, any]):
        for callback in self._on_options_change.values():
            callback(self, options_changed)

    def on_options_change(self, callback: LayerOptionsChangeDispatcher):
        uuid = uuid4().hex
        while uuid in self._on_options_change:
            uuid = uuid4().hex
        self._on_options_change[uuid] = callback
        return DispatcherUnbind(self._on_options_change, uuid)

    def _ipython_display_(self):
        from ..view2d import View2D  # noqa: I001
        from IPython.core.display import display

        return display(View2D(self))


# ======================================================================================================================
#   Layer base class list
# ======================================================================================================================
class LayersList(metaclass=abc.ABCMeta):
    _layers: dict[str, Layer]
    _layers_alias: dict[str, str]
    _layers_binding: dict[str, list[DispatcherUnbind]]
    _main_layer: str | None

    def __init__(self):
        super(LayersList, self).__init__()
        self._layers = {}
        self._layers_alias = {}
        self._layers_binding = {}
        self._update_lock = ContextLock(self.__release_update_lock)
        self._main_layer = None

    # --- Public methods to add, manipulate and remove layers ---
    def add_layer(self, layer: Layer, alias: str | None = None, domain: LayerDomain | None = None):
        """
        Add a layer to the list, placing it on top of the other (if no z_index is specified).
        If the layer is already in the list, it will be duplicated.
        If the alias is already in the list, the old layer will be removed.
        If the domain is not specified, the layer will be fitted to the main layer domain.

        :param layer: The layer to add.
        :param alias: The alias to use for the layer.
        :param domain: The domain to use for the layer. Can be a Rect or a fit indication:
            'fit_width', 'fit_height', 'fit_inner', 'fit_outer'. (By default: 'fit_width')

        """
        main_layer = False
        if alias is None:
            for i in range(1, len(self._layers) + 2):
                alias = f"{layer.layer_type.title()} {i:02d}"
                if alias not in self._layers_alias:
                    break
        elif alias in self:
            main_layer = self[alias].uuid == self._main_layer
            self.remove_layer(alias)

        if layer.label is None:
            layer.label = alias

        if layer.uuid in self._layers:
            layer = layer.duplicate()

        if layer.z_index == -1:
            layer.z_index = max([layer.z_index for layer in self] + [0]) + 1

        if domain is None:
            domain = layer.domain if layer.domain_mode == "manual" else layer.domain_mode

        if self._main_layer is None:
            # If no main layer is set, the first layer added will be the main layer
            if not Rect.is_rect(domain):
                domain = Rect.from_size(layer.shape)
            layer.domain = domain
            main_layer = True
        else:
            layer.set_main_shape(self.main_layer.domain, FitMode.WIDTH if domain is None else domain)

        self._layers[layer.uuid] = layer
        self._layers_alias[alias] = layer.uuid
        self._bind_layer(layer)
        self._send_new_layers([layer])

        if main_layer:
            self.main_layer = layer

    def remove_layer(self, layer: str | Layer | int):
        layer = self._item_to_layer(layer)
        if layer.uuid == self._main_layer:
            self.main_layer = next((layer for layer in self if layer.uuid != self._main_layer), None)

        self._send_delete_layers([layer])
        self._unbind_layer(layer)
        del self._layers_alias[self.get_layers_alias(layer)]
        del self._layers[layer.uuid]

    def update_all_options(self, options, layer_selector: str | Iterable[str | Layer] | LayerSelector | None):
        layers = self.get_layers(layer_selector)
        with self._update_lock:
            for layer in layers:
                layer.set_options(options)

    def update_options(self, layers_options: Mapping[str | Layer, Mapping[str, any]]):
        layers_options = {self[k] if isinstance(k, str) else k: v for k, v in layers_options.items()}
        with self._update_lock:
            if self.main_layer in layers_options and "domain" in layers_options[self.main_layer]:
                self._update_main_layer_domain()
            for layer, opt in layers_options.items():
                layer.set_options(opt)
        # self._send_update_layers_options(layers_options)

    def update_all_data(self, data: any, layer_selector: str | Iterable[str] | LayerSelector | None):
        layers = self.get_layers(layer_selector)
        with self._update_lock:
            for layer in layers:
                layer.update_data(data)

    def update_data(self, data: Mapping[str | Layer, any]):
        data = {self[k] if isinstance(k, str) else k: v for k, v in data.items()}
        with self._update_lock:
            for layer, d in data.items():
                layer.update_data(d)

    # --- Accessors for layers and aliases ---
    @property
    def layers(self) -> dict[str, Layer]:
        return {alias: self._layers[uuid] for alias, uuid in self._layers_alias.items()}

    @property
    def layers_alias(self) -> tuple[str]:
        return tuple(self._layers_alias.keys())

    @property
    def main_layer(self) -> Layer | None:
        return self._layers[self._main_layer] if self._main_layer is not None else None

    @main_layer.setter
    def main_layer(self, main_layer: str | Layer | None):
        if main_layer is None:
            self._main_layer = None
        else:
            main_layer = self._item_to_layer(main_layer)
            new_domain = main_layer.domain

            if self._main_layer is not None:
                previous_domain = self.main_layer.domain
                transform = Transform.from_rects(previous_domain, new_domain)
            else:
                transform = None

            self._main_layer = main_layer.uuid
            self._update_main_layer_domain(transform)

    def _update_main_layer_domain(self, transform: Transform | None = None):
        with self._update_lock:
            new_domain = self.main_layer.domain
            for layer in self:
                if layer.uuid != self._main_layer:
                    layer.set_main_shape(new_domain, transform)

    def get_layers(
        self,
        layers_selector: str | Layer | Iterable[str | Layer] | LayerSelector | None,
        only_visible=False,
        sort_zindex=False,
        layer_type: Type[Layer] | str | Iterable[Type[Layer] | str] | None = None,
    ) -> list[Layer]:
        match layers_selector:
            case None:
                layers = self._layers.values()
            case str():
                try:
                    layers = [self[layers_selector]]
                except KeyError:
                    raise ValueError(f"Unknown layer {layers_selector}") from None
            case Layer():
                if layers_selector not in self:
                    raise ValueError("The provided layer is not in the list.")
                layers = [layers_selector]
            case LayerSelector():
                layers = [layer for name, layer in self.items() if layers_selector(name, layer)]
            case _:  # Iterable[str | Layer]
                layers = [self[layer] for layer in layers_selector]

        if only_visible:
            layers = [layer for layer in layers if layer.visible]
        if layer_type is not None:
            if not isinstance(layer_type, tuple):
                layer_type = (layer_type,)
            layers = [
                layer
                for layer in layers
                if any(
                    layer.layer_type == l_type if isinstance(l_type, str) else isinstance(layer, l_type)
                    for l_type in layer_type
                )
            ]
        if sort_zindex:
            layers.sort(key=lambda layer: layer.z_index)
        return layers

    def get_layers_alias(
        self,
        layers: Layer | Iterable[Layer] | None = None,
        sort_zindex=False,
        only_visible=False,
        by_type: Type[Layer] | str | Iterable[Type[Layer] | str] | None = None,
    ) -> str | list[str]:
        single_layer = isinstance(layers, Layer)
        layers = self.get_layers(layers, sort_zindex=sort_zindex, only_visible=only_visible, layer_type=by_type)

        layers_alias = []
        for layer in layers:
            for alias, uuid in self._layers_alias.items():
                if uuid == layer.uuid:
                    layers_alias.append(alias)
                    break
            else:
                layers_alias.append(None)
        return layers_alias[0] if single_layer else layers_alias

    def layers_domain(self) -> Rect:
        domain = Rect.empty()
        for layer in self:
            domain = domain | Rect.from_size(layer.shape) if layer.domain == "auto" else layer.domain
        return domain

    # --- Item and Iterables accessors ---
    def __len__(self):
        return len(self._layers)

    def __iter__(self):
        return self._layers.values().__iter__()

    def __contains__(self, item: str | Layer):
        match item:
            case str():
                return item in self._layers_alias
            case Layer():
                return item.uuid in self._layers

    def __getitem__(self, key: int | str) -> Layer:
        return self._item_to_layer(key)

    def __setitem__(self, key: str, value: Layer | any):
        if isinstance(value, Layer):
            self.add_layer(value, key)
        else:
            self[key].update_data(value)

    def __delitem__(self, key: int | str):
        self.remove_layer(key)

    def aliases(self):
        return self._layers_alias.keys()

    def values(self):
        return self._layers.values()

    def items(self):
        return zip(self._layers_alias.keys(), self._layers.values(), strict=True)

    # --- Private methods for layer handling ---
    def _item_to_layer(self, item: int | str | Layer) -> Layer:
        if isinstance(item, (int, slice)):
            layers = list(self._layers.values())
            return layers[item]
        elif isinstance(item, str):
            if item not in self._layers_alias:
                raise KeyError(f"No layer named {item}.")
            return self._layers[self._layers_alias[item]]
        elif isinstance(item, Layer):
            if item not in self:
                raise KeyError("This layer is not part of the list.")
            return item

    # --- Private methods for communication ---
    def _bind_layer(self, layer: Layer):
        # Bind options events
        self._layers_binding[layer.uuid] = [
            layer.on_data_change(self.__update_layer_data),
            layer.on_options_change(self.__update_layer_options),
        ]

    def _unbind_layer(self, layer: Layer):
        for unbind in self._layers_binding.get(layer.uuid, ()):
            unbind()

    def __update_layer_data(self, layer: Layer):
        if not self._update_lock:
            self._send_update_layers_data([layer])
        else:
            self._update_lock.add("data", layer)

    def __update_layer_options(self, layer: Layer, options: Mapping[str, any]):
        if not self._update_lock:
            self._send_update_layers_options({layer: options})
        else:
            self._update_lock.add("options", layer)

    def __release_update_lock(self, updated: dict[str, set]):
        if "data" in updated:
            self._send_update_layers_data(updated["data"])
        elif "options" in self._update_lock:
            self._send_update_layers_options({layer: layer.options for layer in updated["options"]})

    # --- Abstract methods for communication ---
    @abc.abstractmethod
    def _send_new_layers(self, layers: Iterable[Layer]):
        ...

    @abc.abstractmethod
    def _send_delete_layers(self, layers: Iterable[Layer]):
        ...

    @abc.abstractmethod
    def _send_update_layers_data(self, layers: Iterable[Layer]):
        ...

    @abc.abstractmethod
    def _send_update_layers_options(self, layers_options: Mapping[Layer, Mapping[str, any]]):
        ...



#!/usr/bin/env python
# coding: utf-8
import json

# Copyright (c) Gabriel Lepetit-Aimon.
# Distributed under the terms of the Modified BSD License.
from typing import Iterator, Mapping, Tuple

import numpy as np
import traitlets

from ._frontend import ABCHasTraitMeta, BaseI3PWidget
from .layers.layer_base import Layer, LayersList
from .layers.layers_2d import LayerGraph, LayerImage, LayerLabel, ValueBound
from .utilities.event import ClickEvent, EventsDispatcher
from .utilities.func import FlagContext


class View2D(LayersList, BaseI3PWidget, metaclass=ABCHasTraitMeta):
    _layers_data = traitlets.Dict(key_trait=traitlets.Unicode(), value_trait=traitlets.Bytes()).tag(
        sync=True,
    )
    _layers_options = traitlets.Dict(key_trait=traitlets.Unicode(), value_trait=traitlets.Unicode()).tag(sync=True)
    _domain = traitlets.Tuple(trait=(int, int, int, int)).tag(sync=True)
    _area_selected = traitlets.Union(
        (traitlets.Tuple(trait=(int, int, int, int)),), allow_none=True, default_value=None
    ).tag(sync=True)
    _hide_foreground = traitlets.Bool(False).tag(sync=True)
    _transform = traitlets.Tuple((0, 0, 1e-8), trait=(float, float, float)).tag(sync=True)
    _target_transform = traitlets.Tuple((0, 0, 1e-8), trait=(float, float, float)).tag(sync=True)
    linkedTransformGroup = traitlets.Unicode(None, allow_none=True).tag(sync=True)

    _model_name = traitlets.Unicode("JView2DModel").tag(sync=True)
    _view_name = traitlets.Unicode("JView2D").tag(sync=True)
    _loading = traitlets.Bool(False).tag(sync=True)

    _left_ruler = traitlets.Bool(True).tag(sync=True)
    _top_ruler = traitlets.Bool(True).tag(sync=True)

    def __init__(self, layers: Iterator[Layer] | Layer = ()):
        super(View2D, self).__init__()
        self.on_click: EventsDispatcher[ClickEvent] = EventsDispatcher()
        self._transmit = FlagContext(self.__set_transmitting)

        if isinstance(layers, Layer):
            layers = (layers,)
        for layer in layers:
            self.add_layer(layer)

    # --- Implementation of LayersList abstract methods---
    def _send_new_layers(self, layers: Iterator[Layer]):
        with self._transmit:
            self._send_update_layers_data(layers)
            self.__send_all_layers_options()

    def _send_delete_layers(self, layers: Iterator[Layer]):
        with self._transmit:
            self.__send_all_layers_options()
            for layer in layers:
                del self._layers_data[self.get_layers_alias(layer)]

    def _send_update_layers_options(self, options: Mapping[str, str]):
        with self._transmit:
            # Ignore diff because the whole dict is sent anyway
            self.__send_all_layers_options()

    def _send_update_layers_data(self, layers: Iterator[Layer]):
        current_data = self._layers_data.copy()
        for layer in layers:
            current_data[self.get_layers_alias(layer)] = layer.get_data().to_json_bytes()
        with self._transmit:
            self._layers_data = current_data

    def __send_all_layers_options(self):
        layers_options = {
            self.get_layers_alias(layer): json.dumps(layer.options, ensure_ascii=False).encode("utf8") for layer in self
        }
        with self._transmit:
            self._layers_options = layers_options
            if self.main_layer:
                self._domain = self.main_layer.domain

    def __set_transmitting(self, value: bool):
        self._loading = value

    # --- Specialized methods to add layers ---
    def add_image(
        self,
        img,
        name: str | None = None,
        vmax: ValueBound = "auto",
        vmin: ValueBound = "auto",
        resize_buffer: Tuple[int, int] | int | None = None,
        options=None,
        **opts,
    ) -> LayerImage:
        layer = LayerImage(img, vmax=vmax, vmin=vmin, resize_buffer=resize_buffer)
        self.add_layer(layer, alias=name)
        opts.update(options or {})
        if opts:
            layer.set_options(opts)
        return layer

    def add_label(
        self, label, name: str | None = None, colormap: str | None = None, options=None, **opts
    ) -> LayerLabel:
        layer = LayerLabel(label, colormap=colormap)
        self.add_layer(layer, alias=name)
        opts.update(options or {})
        if opts:
            layer.set_options(opts)
        return layer

    def add_graph(
        self,
        adjacency_list: np.ndarray,
        nodes_yx: tuple[np.ndarray, np.ndarray],
        edge_labels: np.ndarray | None = None,
        name: str | None = None,
        options=None,
        **opts,
    ) -> LayerGraph:
        layer = LayerGraph(adjacency_list, nodes_yx, edge_labels)
        self.add_layer(layer, alias=name)
        opts.update(options or {})
        if opts:
            layer.set_options(opts)
        return layer

    # --- Events handling ---
    def on_events(self, event, data):
        match event:
            case "onclick":
                self.on_click.dispatch(
                    ClickEvent(x=data["x"], y=data["y"], button=data["button"], modifiers=data["modifiers"])
                )

    def wait_click(self):
        return self.on_click.create_future()

    def goto(self, pos, scale=None):
        if scale is None:
            scale = self._transform[-1]
        self._target_transform = (pos[0], pos[1], scale)

    @property
    def transform(self):
        return self._transform


def imshow(image, vmax: ValueBound = "auto", vmin: ValueBound = "auto", name: str | None = "background"):
    viewer = View2D()
    viewer.add_image(image, vmax=vmax, vmin=vmin, name=name)
    return viewer


def sync_views(*views: View2D):
    import uuid

    sync_uuid = uuid.uuid4().hex
    for view in views:
        view.linkedTransformGroup = sync_uuid



from typing import Callable, List


def colormap_by_name(name="catppuccin") -> List[str]:
    assert isinstance(name, str), f"Invalid colormap name {name}. Must be str."
    match name:
        case "catppuccin":
            return [
                "#8caaee",
                "#99d1db",
                "#a6d189",
                "#ef9f76",
                "#e78284",
                "#f4b8e4",
                "#f2d5cf",
                "#babbf1",
                "#85c1dc",
                "#81c8be",
                "#e5c890",
                "#ea999c",
                "#ca9ee6",
                "#eebebe",
            ]
        case _:
            return [check_color(name)]


def check_color(color: str) -> str:
    import re

    if re.match(r"^#(?:[0-9a-fA-F]{3,4}){1,2}$", color) is not None:
        return color
    else:
        import webcolors

        try:
            return webcolors.name_to_hex(color)
        except ValueError:
            raise ValueError(f"Invalid color name {color}.") from None


class ColorRange(dict):
    def __init__(self, colors: dict[float, str | tuple[str, str]], update_callback: Callable = None):
        super().__init__()
        for i, c in colors.items():
            if isinstance(c, str):
                self[i] = check_color(c)
            elif isinstance(c, tuple) and len(c) == 2:
                self[i] = tuple(check_color(c) for c in c)
            else:
                raise ValueError(f"Invalid color {c}.")
        self._update_callbacks = [update_callback]



#!/usr/bin/env python
# coding: utf-8
# Copyright (c) Gabriel Lepetit-Aimon.
# Distributed under the terms of the Modified BSD License.
from __future__ import annotations

import abc

import traitlets
from IPython.display import display
from ipywidgets import DOMWidget
from traitlets import Int, Unicode

from ._version import __version__

"""
Information about the frontend package of the ipywidgets.
"""
module_name = "jppype"
module_version = __version__


class BaseI3PWidget(DOMWidget):
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _model_name = Unicode(f"MODEL_NAME").tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode(f"VIEW_NAME").tag(sync=True)
    _instance_id = Int(0).tag(sync=True)

    __last_instance_id = 0

    def __init__(self):
        self._instance_id = BaseI3PWidget.__last_instance_id
        BaseI3PWidget.__last_instance_id += 1
        super(BaseI3PWidget, self).__init__()
        self.on_msg(self._on_custom_msg_received)

    def _on_custom_msg_received(self, widget, content, buffer):
        if content.get("event", None) and isinstance(content.get("data", None), dict):
            self.on_events(content["event"], content["data"])

    def on_events(self, event, data):
        pass

    def show(self):
        return display(self)


class ABCHasTraitMeta(abc.ABCMeta, traitlets.MetaHasTraits):
    pass



from __future__ import annotations

__all__ = ["LayerImage", "LayerLabel", "LayerIntensityMap", "LayerGraph", "LayerQuiver"]

import base64
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Literal, Tuple, TypeAlias

import cv2
import numpy as np

from ..utilities.color import check_color
from ..utilities.geometric import Rect, Transform
from ..utilities.image import fit_resize, load_image_from_url
from .layer_base import Layer, LayerData, LayerDomain

if TYPE_CHECKING:
    from torch import Tensor

    ImageDataType = str | Path | np.ndarray | Tensor

ValueBound: TypeAlias = Literal["auto"] | float | None


class LayerImage(Layer):
    def __init__(
        self,
        image: ImageDataType,
        alpha: ImageDataType | None = None,
        vmax: ValueBound = "auto",
        vmin: ValueBound = "auto",
        resize_buffer: Tuple[int, int] | int | None = None,
    ):
        super().__init__("image")
        self.buffer_size = resize_buffer
        self._image = None
        self._image_alpha = None
        self._alpha = None

        self.image = image
        self.alpha = alpha

        self.vmin = vmin
        self.vmax = vmax

    # --- Properties ---
    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, img: ImageDataType):
        if isinstance(img, int | float | bool):
            self._image = np.empty(self.shape[:2], dtype=np.float).fill(img)
            self._image_alpha = None
        else:
            img = LayerImage.load_data(img, self.buffer_size)
            if len(img.shape) == 3 and img.shape[2] == 4:
                self._image_alpha = img[:, :, 3]
                self._image = img[:, :, :3]
            else:
                self._image = img
                self._image_alpha = None
        self._notify_data_change()
        self._notify_options_change({"domain": self.shape})

    @property
    def alpha(self):
        return self._alpha if self._alpha is not None else self._image_alpha

    @alpha.setter
    def alpha(self, map: ImageDataType | None):
        if map is None:
            self._alpha = None
        else:
            self._alpha = LayerIntensityMap.load_data(map, self.buffer_size)
        self._notify_data_change()

    # --- Implementation of layer's abstract methods ---
    def _fetch_data(self, resize: tuple[int, int] | None = None) -> LayerData:
        img = self._image

        if self.vmax == "auto":
            vmax = np.max(img)
        else:
            vmax = self.vmax

        if self.vmin == "auto":
            vmin = np.min(img)
        else:
            vmin = self.vmin

        if self.vmin == "auto" and self.vmax == "auto":
            if vmin < 0 < vmax and 2 / 3 < -vmin / vmax < 3 / 2:
                vmax = max(-vmin, vmax)
                vmin = -vmax
            elif vmax <= 1 and vmin >= 0:
                vmax = 1
                vmin = 0
            elif vmax > 0 and abs(vmin) < vmax * 0.1:
                vmin = 0
            elif vmin < 0 and abs(vmax) < abs(vmin) * 0.1:
                vmax = 0

        elif self.vmax == "auto":
            if vmin < 0 and abs(vmax) < abs(vmin) * 0.1:
                vmax = 0
            elif vmax <= 1 and vmin >= 0:
                vmax = 1
        elif self.vmin == "auto":
            if vmax > 0 and abs(vmin) < vmax * 0.1:
                vmin = 0

        if vmin is not None:
            img = img - np.min(img)
            if vmax is not None:
                vmax = vmax - vmin
        if vmax is not None:
            img = img / vmax * 255.0

        alpha = self.alpha
        h, w = img.shape[:2]
        if resize is not None:
            img = fit_resize(img, resize)
            if alpha is not None:
                alpha = fit_resize(alpha, resize, interpolation=cv2.INTER_NEAREST)

        if alpha is not None:
            img = np.concatenate((img, alpha[..., None]), axis=-1)
            img_format = "png"
        else:
            img_format = "jpg"

        return LayerData(self.encode_url(img, img_format), infos=dict(width=w, height=h))

    def _shape(self):
        return self._image.shape[:2] if self._image is not None else (0, 0)

    def _fetch_item(self, x: int, y: int) -> dict:
        return {"value": self.image[y, x]}

    def _fetch_graphs(self, rect: Tuple[float, float], **kwargs) -> Dict[str, any]:
        return {}

    def update_data(self, data: any):
        self.image = data

    # --- Utilities static methods ---
    @staticmethod
    def load_data(img: ImageDataType, resize=None) -> np.ndarray:
        """
        Cast image to numpy array with shape (H, W, C) or (H, W).

        Parameters
        ----------
        img: np.ndarray or torch.Tensor

        resize: tuple[int, int] or int or None

        Returns
        -------
        Casted image

        """
        if isinstance(img, (str, Path)):
            img = load_image_from_url(img)
        elif type(img).__qualname__ == "Tensor":
            img = img.detach().cpu().numpy()

        if len(img.shape) == 3:
            if img.shape[0] in (1, 3, 4) and img.shape[2] not in (1, 3, 4):
                img = img.transpose((1, 2, 0))
            if img.shape[2] == 1:
                img = img[:, :, 0]
        elif len(img.shape) != 2:
            raise ValueError(f"Invalid image shape {img.shape}, must be (H, W, C) or (H, W).")

        if resize is not None:
            img = fit_resize(img, resize)

        return img

    @staticmethod
    def encode_url(img: np.ndarray, format="jpg") -> str:
        _, data = cv2.imencode("." + format, img)
        return f"data:image/{format};base64," + base64.b64encode(data).decode("ascii")


class LayerLabel(Layer):
    def __init__(self, label_map, colormap: Dict[int, str] | List[str] | str | None = None):
        super().__init__("label")
        self.label_map = label_map
        self.colormap = colormap
        self.foreground = True

    @property
    def label_map(self):
        return self._label_map

    @label_map.setter
    def label_map(self, data):
        data = LayerLabel.load_data(data)
        self._label_map = data
        self._notify_data_change()

    @property
    def colormap(self):
        """Get colormap as dict[int, str]."""
        return self._options.get("cmap", None)

    @colormap.setter
    def colormap(self, cmap):
        cmap = self.check_label_colormap(cmap)
        self._options["cmap"] = cmap
        self._notify_options_change({"cmap": cmap})

    def _fetch_data(self, resize: Tuple[int, int] | None = None) -> LayerData:
        label = self._label_map

        h, w = label.shape[:2]
        if resize is not None:
            label = fit_resize(label, resize, interpolation=cv2.INTER_NEAREST)
        labels = np.unique(label).tolist()

        return LayerData(
            LayerLabel.encode_url(label),
            infos={"width": w, "height": h, "labels": labels},
        )

    def _shape(self):
        return self.label_map.shape if self.label is not None else (0, 0)

    def _fetch_item(self, x: int, y: int) -> dict:
        return {"value": self._label[y, x]}

    def _fetch_graphs(self, rect: Tuple[float, float], **kwargs) -> Dict[str, any]:
        return {}

    def update_data(self, data: any):
        self.label_map = data

    # --- Utilities static methods ---
    @staticmethod
    def load_data(data, resize=None) -> np.ndarray:
        """
        Cast label map (2d map of integer) to a numpy array with shape (H, W).

        Parameters
        ----------
        data: np.ndarray or torch.Tensor or path to a label map

        resize: tuple[int, int] or int or None

        Returns
        -------
        Casted image

        """
        if isinstance(data, str):
            data = load_image_from_url(data)
            if data.ndim == 3:
                data = data.mean(axis=-1).astype(np.uint8)
        elif type(data).__qualname__ == "Tensor":
            data = data.detach().cpu().numpy()

        assert isinstance(data, np.ndarray), f"Invalid label type {type(data)}. Must be numpy.ndarray."
        assert data.ndim == 2, f"Invalid label shape {data.shape}. Must be (H, W)."

        error = ValueError(f"Invalid label type {data.dtype}. Must be positive integer encoded on maximum 32 bits.")
        if data.dtype.kind not in "?bBiu":
            raise error
        elif np.min(data) < 0 or np.max(data) >= 2**32:
            raise error

        if resize is not None:
            data = fit_resize(data, resize, interpolation=cv2.INTER_NEAREST)

        return data.astype(np.uint32)

    @staticmethod
    def encode_url(label: np.ndarray) -> str:
        alpha = (label >> 24).astype(np.uint8)
        red = ((label >> 16) & 0xFF).astype(np.uint8)
        green = ((label >> 8) & 0xFF).astype(np.uint8)
        blue = (label & 0xFF).astype(np.uint8)

        label = np.stack((red, green, blue, 255 - alpha), axis=-1)
        return LayerImage.encode_url(label, "png")

    @staticmethod
    def check_label_colormap(cmap, null_label=True) -> Dict[int, str]:
        from ..utilities.color import check_color, colormap_by_name

        if not null_label and isinstance(cmap, dict):
            if None in cmap:
                cmap[-1] = cmap.pop(None)
            cmap = {k + 1: v for k, v in cmap.items()}
        match cmap:
            case list() | tuple():
                cmap = {0: list(cmap)}
            case str():
                cmap = {0: colormap_by_name(cmap)}
            case None:
                cmap = {0: colormap_by_name()}
        if not isinstance(cmap, dict) or any(
            not isinstance(k, int) or not isinstance(v, str) for k, v in cmap.items() if k != 0
        ):
            raise ValueError(f"Invalid colormap {cmap}. Must be dict[int, str].")

        return {int(k): check_color(v) if k != 0 else [check_color(_) for _ in v] for k, v in cmap.items()}


class LayerIntensityMap(Layer):
    def __init__(self, map: np.ndarray, color_range):
        from ..utilities.color import ColorRange

        super().__init__("intensity")
        self.map = map
        self._color_range = ColorRange(color_range)

    @property
    def map(self):
        return self._map

    @map.setter
    def map(self, map):
        map = LayerIntensityMap.load_data(map)
        self._map = map
        self._notify_data_change()

    @property
    def color_range(self):
        return self._color_range

    @color_range.setter
    def color_range(self, color_range):
        assert isinstance(color_range, list) and len(color_range) == 2, f"Invalid color range {color_range}."
        self._options["color_range"] = color_range
        self._notify_options_change({"color_range": color_range})

    def _fetch_data(self, resize: Tuple[int, int] | None = None) -> LayerData:
        map = self.map
        if resize is not None:
            map = fit_resize(map, resize, interpolation=cv2.INTER_NEAREST)
        return LayerData(LayerIntensityMap.encode_url(map))

    def _shape(self):
        return self.map.shape if self.map is not None else (0, 0)

    def _fetch_item(self, x: int, y: int) -> dict:
        return {"value": self._map[y, x]}

    # --- Utilities static methods ---
    @staticmethod
    def load_data(data, resize=None) -> np.ndarray:
        """
        Cast intensity map (2d map of float) to a numpy array with shape (H, W).

        Parameters
        ----------
        data: np.ndarray or torch.Tensor or path to a label map

        resize: tuple[int, int] or int or None

        Returns
        -------
        Casted image

        """

        if isinstance(data, str):
            data = load_image_from_url(data)
        elif type(data).__qualname__ == "Tensor":
            data = data.detach().cpu().numpy()

        assert isinstance(map, np.ndarray), f"Invalid map type {type(map)}. Must be numpy.ndarray."
        assert map.ndim == 2, f"Invalid map shape {map.shape}. Must be (H, W)."

        if resize is not None:
            data = fit_resize(data, resize)

        return data.astype(np.float32)

    @staticmethod
    def encode_url(label: np.ndarray) -> str:
        label = label.astype(np.float32).tobytes()
        alpha = (label >> 24).astype(np.uint8)
        red = ((label >> 16) & 0xFF).astype(np.uint8)
        green = ((label >> 8) & 0xFF).astype(np.uint8)
        blue = (label & 0xFF).astype(np.uint8)

        label = np.stack((red, green, blue, 255 - alpha), axis=-1)
        return LayerImage.encode_url(label, "png")


class LayerGraph(Layer):
    def __init__(
        self, adjacency_list, nodes_coordinates, edge_map=None, nodes_domain=None, nodes_cmap=None, edges_cmap=None
    ):
        super().__init__("graph")
        self.set_graph(adjacency_list, nodes_coordinates, edge_map, nodes_domain)
        self.nodes_cmap = nodes_cmap
        self.edges_cmap = edges_cmap
        self.edges_opacity = 0.7
        self.node_labels_visible = False
        self.edge_labels_visible = False
        self.edge_map_visible = edge_map is None
        self.foreground = True

    def set_graph(self, adjacency_list, nodes_coordinates, edge_map=None, nodes_domains: Rect | None = None):
        if nodes_domains is None and edge_map is not None:
            nodes_domains = Rect.from_size(edge_map.shape)
        self._set_adjacency_list(adjacency_list, check_dim=False)
        self._set_nodes_coordinates(nodes_coordinates, nodes_domains)
        self._set_edge_map(edge_map)
        self._notify_data_change()

    def set_options(self, options: Dict[str, any], raise_on_error: bool = True):
        for k, v in options.items():
            if k == "nodes_cmap":
                self._options[k] = LayerLabel.check_label_colormap(v, null_label=False)
            elif k == "edges_cmap":
                self._options[k] = LayerLabel.check_label_colormap(v, null_label=False)
            elif k == "edges_opacity":
                self._options[k] = min(max(float(v), 0), 1)
            elif k == "node_labels_visible":
                self._options[k] = bool(v)
            elif k == "edge_labels_visible":
                self._options[k] = bool(v)
            elif k == "edge_map_visible":
                self._options[k] = bool(v)
        super().set_options(options, raise_on_error)

    @property
    def adjacency_list(self):
        return getattr(self, "_adjacency_list", None)

    @adjacency_list.setter
    def adjacency_list(self, data):
        self._set_adjacency_list(data)
        self._notify_data_change()

    def _set_adjacency_list(self, adj, check_dim=True):
        if type(adj).__qualname__ == "Tensor":
            adj = adj.detach().cpu().numpy()
        adj = np.asarray(adj)
        assert adj.ndim == 2 and adj.shape[1] == 2, f"Invalid adjacency list shape {adj.shape}. Must be (E, 2)."
        if check_dim:
            if self.nodes_coordinates is not None:
                assert (
                    adj.max() < self.nodes_coordinates.shape[0]
                ), f"Invalid adjacency list. {self.nodes_coordinates.shape[0]} nodes are expected."
            if self.edge_map is not None:
                assert (
                    adj.shape[0] == self.edge_map.max()
                ), f"Invalid adjacency list. {self.edge_map.max()} edges are expected."
        self._adjacency_list = adj.astype(np.uint32)

    @property
    def nodes_coordinates(self) -> np.ndarray:
        return getattr(self, "_nodes_coordinates", None)

    @nodes_coordinates.setter
    def nodes_coordinates(self, data):
        self._set_nodes_coordinates(data)
        self._notify_data_change()

    def _set_nodes_coordinates(self, node_yx, nodes_domain=None, check_dim=True):
        node_yx = np.asarray(node_yx)
        assert (
            node_yx.ndim == 2 and node_yx.shape[1] == 2
        ), f"Invalid  nodes coordinates shape {node_yx.shape}. Must be (nbNodes, 2)."
        if check_dim:
            if self.adjacency_list is not None:
                assert node_yx.shape[0] > self.adjacency_list.max(), (
                    f"Invalid nodes coordinates shape {node_yx.shape}. "
                    f"Expected at least {self.adjacency_list.max()+1} nodes but got {node_yx.shape[0]}."
                )
        if nodes_domain is None:
            if self.edge_map is not None:
                nodes_domain = Rect.from_size(self._edge_map.shape)
            elif not Rect.is_empty(self._main_domain):
                nodes_domain = self._main_domain
        self._nodes_coordinates = node_yx.astype(np.uint32)
        self._nodes_domain = nodes_domain

    @property
    def edge_map(self) -> np.ndarray | None:
        return getattr(self, "_edge_map", None)

    @edge_map.setter
    def edge_map(self, data):
        self._set_edge_map(data)
        self._notify_data_change()

    def _set_edge_map(self, edge_label, check_dim=True):
        if type(edge_label).__qualname__ == "Tensor":
            edge_label = edge_label.detach().cpu().numpy()
        if edge_label is not None:
            error = ValueError(
                f"Invalid edge map type {edge_label.dtype}. " f"Must be positive integer encoded on maximum 32 bits."
            )
            if edge_label.ndim != 2:
                raise error
            elif edge_label.dtype.kind not in "?bBiu":
                raise error
            elif np.min(edge_label) < 0 or np.max(edge_label) >= 2**32:
                raise error
            if check_dim:
                if self.adjacency_list is not None:
                    assert edge_label.max() == self.adjacency_list.shape[0], (
                        f"Invalid edge label: maximum label is {edge_label.max()} "
                        f"but adjacency list contains{self.adjacency_list.shape[0]} edges."
                    )
            self._edge_map = edge_label.astype(np.uint32)
            if self._nodes_domain is None:
                self._nodes_domain = Rect.from_size(self._edge_map.shape)
        else:
            self._edge_map = None

    @property
    def nodes_cmap(self):
        """Get colormap as dict[int, str]."""
        return self._options.get("nodes_cmap", None)

    @nodes_cmap.setter
    def nodes_cmap(self, cmap):
        self.set_options({"nodes_cmap": cmap})

    @property
    def edges_cmap(self):
        """Get colormap as dict[int, str]."""
        return self._options.get("edges_cmap", None)

    @edges_cmap.setter
    def edges_cmap(self, cmap):
        self.set_options({"edges_cmap": cmap})

    @property
    def node_labels_visible(self):
        return self._options.get("node_labels_visible", False)

    @node_labels_visible.setter
    def node_labels_visible(self, cmap):
        self.set_options({"node_labels_visible": cmap})

    @property
    def edge_labels_visible(self):
        return self._options.get("edge_labels_visible", False)

    @edge_labels_visible.setter
    def edge_labels_visible(self, cmap):
        self.set_options({"edge_labels_visible": cmap})

    @property
    def edge_map_visible(self):
        return self._options.get("edge_map_visible", False)

    @edge_map_visible.setter
    def edge_map_visible(self, cmap):
        self.set_options({"edge_map_visible": cmap})

    @property
    def edges_opacity(self):
        return self._options.get("edges_opacity", 1.0)

    @edges_opacity.setter
    def edges_opacity(self, opacity):
        self.set_options({"edges_opacity": opacity})

    def set_main_shape(self, main_domain: Rect, transform_domain: Transform | LayerDomain | None = None):
        if self._nodes_domain is None:
            self._nodes_domain = main_domain
        super().set_main_shape(main_domain, transform_domain)

    def _fetch_data(self, resize: Tuple[int, int] | None = None) -> LayerData:
        data = dict(
            adj=self._adjacency_list.astype(int).tolist(), nodes_yx=self.nodes_coordinates.astype(float).tolist()
        )
        if self.edge_map is not None:
            data["edgeMap"] = LayerLabel.encode_url(self.edge_map)
        return LayerData(
            data=data,
            infos={
                "nbNodes": int(self._adjacency_list.max()) + 1,
                "nodesDomain": self._nodes_domain,
            },
        )

    def _shape(self):
        return self.edge_map.shape if self.edge_map is not None else self._nodes_domain

    def _fetch_item(self, x: int, y: int) -> dict:
        return {"value": self.edge_map[y, x]}

    def _fetch_graphs(self, rect: Tuple[float, float], **kwargs) -> Dict[str, any]:
        return {}

    def update_data(self, data: any):
        if isinstance(data, tuple) and len(data) in (2, 1):
            self.set_graph(*data)
        else:
            self.adjacency_list = data


class LayerQuiver(Layer):
    def __init__(
        self,
        xy,
        uv,
        domain: Rect,
        zoom_scaling: bool = False,
    ):
        super().__init__("quiver")
        self.set_data(xy, uv, domain)
        self.foreground = True
        self.set_options(
            dict(
                color="#333",
                width=2,
                zoom_scaling=zoom_scaling,
            )
        )

    def set_data(self, xy, uv, domain: Rect):
        assert xy.ndim == 2 and xy.shape[1] == 2, f"Invalid arrow_xy shape {xy.shape}. Must be (nbArrows, 2)."
        assert uv.ndim == 2 and uv.shape[1] == 2, f"Invalid arrow_uv shape {uv.shape}. Must be (nbArrows, 2)."
        assert xy.shape[0] == uv.shape[0], f"Invalid arrow_xy and arrow_uv shapes {xy.shape} and {uv.shape}."
        self._arrows_xy_uv = np.concatenate((xy, uv), axis=1)
        self._quiver_domain = domain
        self._notify_data_change()

    def set_options(self, options: Dict[str, any], raise_on_error: bool = True):
        for k, v in options.items():
            if k == "zoom_scaling":
                if v not in ("scene", "view", "view_log", "view_sqrt"):
                    if raise_on_error:
                        raise ValueError(f"Invalid zoom_scaling {v}.")
                else:
                    self._options[k] = v
            elif k == "color":
                try:
                    self._options[k] = check_color(v)
                except ValueError:
                    if raise_on_error:
                        raise ValueError(f"Invalid color {v}.") from None
            elif k == "width":
                self._options[k] = max(float(v), 1)
        super().set_options(options, raise_on_error)

    @property
    def zoom_scaling(self):
        return self._options.get("zoom_scaling", False)

    @zoom_scaling.setter
    def zoom_scaling(self, zoom_scaling):
        self.set_options({"zoom_scaling": zoom_scaling})

    @property
    def color(self):
        return self._options.get("color", "#333")

    @color.setter
    def color(self, color):
        self.set_options({"color": color})

    @property
    def width(self):
        return self._options.get("width", 2)

    @width.setter
    def width(self, width):
        self.set_options({"width": width})

    def _fetch_data(self, resize: Tuple[int, int] | None = None) -> LayerData:
        data = dict(arrows=self._arrows_xy_uv.astype(float).tolist())

        return LayerData(
            data=data,
            infos={
                "quiverDomain": self._quiver_domain,
            },
        )

    def _shape(self) -> Tuple[int]:
        return self._quiver_domain

    def _fetch_item(self, x: int, y: int) -> dict:
        return {}

    def _fetch_graphs(self, rect: Tuple[float, float], **kwargs) -> Dict[str, any]:
        return {}

    def update_data(self, data: any):
        match data:
            case (np.ndarray, np.ndarray):
                self.set_data(*data, self._quiver_domain)
            case np.ndarray:
                if data.ndim == 2:
                    self.set_data(self._arrows_xy_uv[:, :2], data, self._quiver_domain)
                elif data.ndim == 4:
                    self.set_data(data[:, :2], data[:, 2:], self._quiver_domain)
